#!/usr/bin/python

# Scans .cpp files for dependencies and produces the Makefile.

import os
import os.path
import re

SOURCE_DIRS = ['src', 'src/external/enet', 'src/external/guichan/src',
               'src/coercri/core', 'src/coercri/enet', 'src/coercri/gcn',
               'src/coercri/gfx', 'src/coercri/network', 'src/coercri/sound',
               'src/coercri/timer',
               'src/coercri/sdl/core', 'src/coercri/sdl/gfx',
               'src/coercri/sdl/sound', 'src/coercri/sdl/timer']

INCLUDE_DIRS = ['src', 'src/misc', 'src/external', 'src/external/guichan/include']



# Strip spaces and remove blank lines from a list of source files
def clean_srcs(srcs):
    srcs = map(lambda x: x.strip(), srcs)
    srcs = filter(lambda x: x != '', srcs)
    return srcs


# Find a file within the INCLUDE_DIRS
def find_include(filename, curdir):
    for i in INCLUDE_DIRS + [curdir]:
        tryfile = os.path.join(i, filename)
        if (os.path.exists(tryfile)):
            return tryfile
    return ""


# Scan a file for dependencies.
all_deps = {}
def scan(filename):
    if filename in all_deps: return
    dependencies = []
    lines = file(filename, "r").readlines()
    for x in lines:
        x = x.strip()
        if (x.startswith('#include')):
            idx = x.find('<')
            if idx == -1: idx = x.find('"')
            idxstart = idx + 1
            idxstop = x.find('>', idxstart)
            if idxstop == -1: idxstop = x.find('"', idxstart)
            included_file = x[idxstart:idxstop]
            where = find_include(included_file, os.path.dirname(filename))
            if where != "":
                dependencies.append(os.path.normpath(where))
    all_deps[filename] = dependencies

# Make sure all recursive dependencies have been found.
def recursion():
    for k in all_deps:
        files = all_deps[k]
        for f in files:
            if not f in all_deps:
                scan(f)
                return True
    return False

def fill_deps():
    while recursion():
        pass

# Extract the list of dependencies for a given file.
def get_deps(filename):
    result = set([filename])
    while True:
        result2 = result
        for f in result:
            immediate_deps = all_deps[f]
            imm_deps_as_set = set(immediate_deps)
            result2 = result2.union(imm_deps_as_set)
        if len(result2) == len(result): break
        result = result2
    result.remove(filename)  # File shouldnt depend on itself.
    return list(result)
            

# Get all source-files in the given dir.
def get_src_files(dir):
    srcs = file(os.path.join(dir, "Sources"), "r").readlines()
    srcs = clean_srcs(srcs)
    return map(lambda x: os.path.join(dir, x), srcs)
       


# Main program

# Get all source files.
srcs = []
for i in SOURCE_DIRS:
    srcs = srcs + get_src_files(i)

# Get a list of object files. Make each object depend on its source file.
objfiles = []
for filename in srcs:
    objname = re.sub(r"\.cpp", r".o", filename)
    objname = re.sub(r"\.c", r".o", objname)
    objfiles.append(objname)
    all_deps[objname] = [filename]

# Recursively get dependencies.
fill_deps()

# Start printing the Makefile.
print """# Makefile for Knights
# This file is generated automatically by makemakefile.py

PREFIX = /usr/local
BIN_DIR = $(PREFIX)/bin
DOC_DIR = $(PREFIX)/share/doc/knights
DATA_DIR = $(PREFIX)/share/knights

BINARY_NAME = knights

CC = gcc
CXX = g++
STRIP = strip

# NOTE: Audio can be disabled by adding -DDISABLE_SOUND to the CFLAGS.
# This is useful if audio doesn't work on your system for some reason.
CFLAGS = -DUSE_FONTCONFIG -DDATA_DIR=$(DATA_DIR) -DNDEBUG -O3 -ffast-math
CXXFLAGS = $(CFLAGS)

INCLUDES = -Isrc -Isrc/misc -Isrc/external -Isrc/external/guichan/include

OWNER = root
GROUP = root
INSTALL = install -o $(OWNER) -g $(GROUP)
"""

print "OFILES =",
for i in objfiles:
    print i,

print """

build: $(BINARY_NAME)

%.o: %.cpp
\t$(CXX) $(CXXFLAGS) `sdl-config --cflags` $(INCLUDES) -c $< -o $@

%.o: %.c
\t$(CC) $(CFLAGS) `sdl-config --cflags` `freetype-config --cflags` $(INCLUDES) -c $< -o $@

$(BINARY_NAME): $(OFILES)
\t$(CXX) $(CXXFLAGS) -o $@ $^ `sdl-config --libs` `freetype-config --libs` -lfontconfig
\t$(STRIP) $@

clean:
\trm -f $(OFILES)
\trm -f $(BINARY_NAME)
"""

# Print the dependencies for each source file.
for obj in objfiles:
    deps = get_deps(obj)
    print obj + ":",
    for d in deps:
        print d,
    print


# Print the 'install' target.
print """
install: install_knights install_docs

install_knights: $(BINARY_NAME)
\t$(INSTALL) -m 755 -d $(BIN_DIR)
\t$(INSTALL) -m 755 -d $(DATA_DIR)
\t$(INSTALL) -m 755 $(BINARY_NAME) $(BIN_DIR)"""

for root, dirs, files in os.walk('knights_data'):
    for f in files:
        print "\t$(INSTALL) -m 644 " + os.path.join(root, f) + " $(DATA_DIR)"

print
print "install_docs:"
print "\t$(INSTALL) -m 755 -d $(DOC_DIR)"
print "\t$(INSTALL) -m 755 -d $(DOC_DIR)/images"

for root, dirs, files in os.walk('docs'):
    for f in files:
        if (f.endswith(".png")):  # slightly crude test but it works for now!
            print "\t$(INSTALL) -m 644 " + os.path.join(root, f) + " $(DOC_DIR)/images"
        else:
            print "\t$(INSTALL) -m 644 " + os.path.join(root, f) + " $(DOC_DIR)"

print

# Print the 'uninstall' target
print """
uninstall:
\trm -f $(BIN_DIR)/$(BINARY_NAME)"""

for root, dirs, files in os.walk('knights_data'):
    for f in files:
        print "\trm -f $(DATA_DIR)/" + os.path.basename(f)

for root, dirs, files in os.walk('docs'):
    for f in files:
        if (f.endswith(".png")):  # see comment above
            print "\trm -f $(DOC_DIR)/images/" + os.path.basename(f)
        else:
            print "\trm -f $(DOC_DIR)/" + os.path.basename(f)
