/*
 * FILE:
 *   cg_listener.cpp
 *
 * AUTHOR:
 *   Stephen Thompson, 2008
 *
 * COPYRIGHT:
 *   Public Domain
 *
 */

#include "cg_graphics.hpp"
#include "cg_input.hpp"
#include "cg_listener.hpp"
#include "../gfx/window.hpp"
#include "../timer/timer.hpp"

#include "guichan.hpp"

namespace Coercri {

    class CGListenerImpl {
    public:
        CGListenerImpl(boost::shared_ptr<Window> win, boost::shared_ptr<gcn::Gui> g, boost::shared_ptr<Timer> tmr)
            : window(win), gui(g), timer(tmr), input(tmr), gui_enabled(false),
              left_shift_pressed(false), right_shift_pressed(false),
              last_mouse_x(0), last_mouse_y(0)
        { }
        
        boost::shared_ptr<Window> window;
        boost::shared_ptr<gcn::Gui> gui;
        boost::shared_ptr<Timer> timer;
        
        CGInput input;
        CGGraphics graphics;
        
        bool gui_enabled;
        bool left_shift_pressed, right_shift_pressed;
        int last_mouse_x, last_mouse_y;
    };

    CGListener::CGListener(boost::shared_ptr<Window> window, boost::shared_ptr<gcn::Gui> gui, boost::shared_ptr<Timer> timer)
        : pimpl(new CGListenerImpl(window, gui, timer))
    {
        pimpl->gui->setInput(&pimpl->input);
        pimpl->gui->setGraphics(&pimpl->graphics);
    }

    void CGListener::enableGui()
    {
        if (pimpl->gui_enabled) return;
        pimpl->gui_enabled = true;

        // Window may have been resized while the GUI was disabled, so do a resize now.
        int w, h;
        pimpl->window->getSize(w, h);
        onResize(w, h);
    }

    void CGListener::disableGui()
    {
        pimpl->gui_enabled = false;
    }

    bool CGListener::isGuiEnabled() const
    {
        return pimpl->gui_enabled;
    }

    bool CGListener::processInput()
    {
        if (!pimpl->gui_enabled) return false;
        
        if (pimpl->input.inputWaiting()) {
            pimpl->gui->logic();
            pimpl->window->invalidateAll();
            return true;
        } else {
            return false;
        }
    }
    
    void CGListener::draw(GfxContext &gc)
    {
        if (!pimpl->gui_enabled) return;

        pimpl->graphics.setTarget(&gc);
        pimpl->gui->draw();
        pimpl->graphics.setTarget(0);
    }

    void CGListener::onResize(int new_width, int new_height)
    {
        if (!pimpl->gui_enabled) return;
        
        if (pimpl->gui->getTop()) {
            pimpl->gui->getTop()->setSize(new_width, new_height);
            pimpl->window->invalidateAll();  // make sure it gets redrawn.
        }
    }

    void CGListener::onRawKeyDown(RawKey rk)
    {
        processRawKey(rk, true);
    }

    void CGListener::onRawKeyUp(RawKey rk)
    {
        processRawKey(rk, false);
    }
    
    void CGListener::processRawKey(RawKey rk, bool down)
    {
        // Tab, Space and Enter have to be handled here since these
        // are interpreted specially by guichan and we need to pass
        // through both the keydown and the keyup events for them to
        // work properly.
        gcn::Key k;
        switch (rk) {
        case RK_LEFT_SHIFT: pimpl->left_shift_pressed = down; return;
        case RK_RIGHT_SHIFT: pimpl->right_shift_pressed = down; return;
        case RK_TAB: k = gcn::Key::TAB; break;
        case RK_SPACE: k = gcn::Key::SPACE; break;
        case RK_KP_ENTER: k = gcn::Key::ENTER; break;
        case RK_RETURN: k = gcn::Key::ENTER; break;
        default: return;
        }

        gcn::KeyInput ki(k, down ? gcn::KeyInput::PRESSED : gcn::KeyInput::RELEASED);
        ki.setShiftPressed(pimpl->left_shift_pressed || pimpl->right_shift_pressed);
        pimpl->input.addKeyInput(ki);
    }

    void CGListener::onCookedKey(CookedKey ck, int character, int mod)
    {
        if (!pimpl->gui_enabled) return;
        
        gcn::Key k;
        if (ck == CK_CHARACTER) {
            if (character == ' ') return;  //SPACE ignored because handled in the raw key event
            else k = character;
        } else {
            switch (ck) {
            //CK_TAB ignored because handled in the raw key event
            //CK_RETURN ignored because handled in the raw key event
            case CK_INSERT: k = gcn::Key::INSERT; break;
            case CK_HOME: k = gcn::Key::HOME; break;
            case CK_PAGE_UP: k = gcn::Key::PAGE_UP; break;
            case CK_DELETE: k = gcn::Key::DELETE; break;
            case CK_END: k = gcn::Key::END; break;
            case CK_PAGE_DOWN: k = gcn::Key::PAGE_DOWN; break;
            case CK_ESCAPE: k = gcn::Key::ESCAPE; break;
            case CK_BACKSPACE: k = gcn::Key::BACKSPACE; break;
            case CK_F1: k = gcn::Key::F1; break;
            case CK_F2: k = gcn::Key::F2; break;
            case CK_F3: k = gcn::Key::F3; break;
            case CK_F4: k = gcn::Key::F4; break;
            case CK_F5: k = gcn::Key::F5; break;
            case CK_F6: k = gcn::Key::F6; break;
            case CK_F7: k = gcn::Key::F7; break;
            case CK_F8: k = gcn::Key::F8; break;
            case CK_F9: k = gcn::Key::F9; break;
            case CK_F10: k = gcn::Key::F10; break;
            case CK_F11: k = gcn::Key::F11; break;
            case CK_F12: k = gcn::Key::F12; break;
            case CK_F13: k = gcn::Key::F13; break;
            case CK_F14: k = gcn::Key::F14; break;
            case CK_F15: k = gcn::Key::F15; break;
            case CK_PRINT: k = gcn::Key::PRINT_SCREEN; break;
            case CK_PAUSE: k = gcn::Key::PAUSE; break;
            case CK_LEFT: k = gcn::Key::LEFT; break;
            case CK_RIGHT: k = gcn::Key::RIGHT; break;
            case CK_UP: k = gcn::Key::UP; break;
            case CK_DOWN: k = gcn::Key::DOWN; break;
            default: return;   // Unknown key
            }
        }

        gcn::KeyInput ki(k, gcn::KeyInput::PRESSED);
        ki.setControlPressed((mod & MOD_CTRL) != 0);
        ki.setAltPressed((mod & MOD_ALT) != 0);
        ki.setMetaPressed((mod & MOD_META) != 0);

        // Note: we don't send the corresponding RELEASED events since
        // Coercri doesn't provide those. In cases where RELEASED is
        // needed, we have to process these through the raw key events
        // (see above).

        pimpl->input.addKeyInput(ki);
    }

    void CGListener::onMouseDown(int x, int y, MouseButton button)
    {
        if (!pimpl->gui_enabled) return;
        
        unsigned int b, t;
        switch (button) {
        case MB_LEFT:
            b = gcn::MouseInput::LEFT;
            t = gcn::MouseInput::PRESSED;
            break;
        case MB_RIGHT:
            b = gcn::MouseInput::RIGHT;
            t = gcn::MouseInput::PRESSED;
            break;
        case MB_MIDDLE:
            b = gcn::MouseInput::MIDDLE;
            t = gcn::MouseInput::PRESSED;
            break;
        case MB_WHEEL_UP:
            b = gcn::MouseInput::EMPTY;
            t = gcn::MouseInput::WHEEL_MOVED_UP;
            break;
        case MB_WHEEL_DOWN:
            b = gcn::MouseInput::EMPTY;
            t = gcn::MouseInput::WHEEL_MOVED_DOWN;
            break;
        default:
            return;
        }
        const int timestamp = pimpl->timer ? pimpl->timer->getMsec() : 0;

        pimpl->input.addMouseInput(gcn::MouseInput(b, t, x, y, timestamp));
        pimpl->last_mouse_x = x;
        pimpl->last_mouse_y = y;
    }

    void CGListener::onMouseUp(int x, int y, MouseButton button)
    {
        if (!pimpl->gui_enabled) return;
        
        unsigned int b;
        switch (button) {
        case MB_LEFT: b = gcn::MouseInput::LEFT; break;
        case MB_RIGHT: b = gcn::MouseInput::RIGHT; break;
        case MB_MIDDLE: b = gcn::MouseInput::MIDDLE; break;
        default: return;
        }
        const int timestamp = pimpl->timer ? pimpl->timer->getMsec() : 0;
        pimpl->input.addMouseInput(gcn::MouseInput(b, gcn::MouseInput::RELEASED, x, y, timestamp));
        pimpl->last_mouse_x = x;
        pimpl->last_mouse_y = y;
    }

    void CGListener::onMouseMove(int x, int y)
    {
        if (!pimpl->gui_enabled) return;
        
        const int timestamp = pimpl->timer ? pimpl->timer->getMsec() : 0;
        pimpl->input.addMouseInput(gcn::MouseInput(gcn::MouseInput::EMPTY, gcn::MouseInput::MOVED, x, y, timestamp));
        pimpl->last_mouse_x = x;
        pimpl->last_mouse_y = y;
    }

    void CGListener::repeatLastMouseInput()
    {
        onMouseMove(pimpl->last_mouse_x, pimpl->last_mouse_y);
    }
}
