/*
 * colour_change.hpp
 *
 * Defines a mapping from old colours to new colours. Used for
 * dynamically changing colours of Graphics, eg for Knight House
 * Colours, or for changing the colour of your potion bottle when you
 * have strength, super etc.
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#ifndef COLOUR_CHANGE_HPP
#define COLOUR_CHANGE_HPP

#include "coercri/gfx/color.hpp"
#include "coercri/network/network_connection.hpp"

#include <vector>
using namespace std;

class ColourChange {
public:
    // Constructor: creates an 'empty' ColourChange object (where all colours are passed
    // through unmodified).
    ColourChange() { }
    
    // empty(): true if this is the 'identity' colour change
    bool empty() const { return mappings.empty(); }
    
    // add: adds a mapping from old_col to new_col
    void add(Coercri::Color old_col, Coercri::Color new_col);

    // lookup: returns false if the src colour is to be left unchanged
    // returns true (and sets new_col) if the src colour is to be
    // changed to new_col.
    bool lookup(Coercri::Color old_col, Coercri::Color &new_col) const;

    // comparison functions.
    bool operator<(const ColourChange &other) const { return mappings < other.mappings; }
    bool operator==(const ColourChange &other) const { return mappings == other.mappings; }

    // Serialization interface
    // Note we don't bother with ID's for colour changes. Instead, the same ColourChange
    // will be written several times if necessary.
    void serialize(Coercri::NetworkConnection &) const;
    explicit ColourChange(Coercri::NetworkConnection &);
    
private:
    // Vector of (old,new) pairs. This is kept sorted (on "old" values) at all times.
    vector<pair<Coercri::Color,Coercri::Color> > mappings;
};

#endif
