/*
 * in_game_screen.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "client_config.hpp"
#include "error_screen.hpp"
#include "game_client.hpp"
#include "game_server.hpp"
#include "gfx_manager.hpp"
#include "gfx_resizer.hpp"
#include "in_game_screen.hpp"
#include "knights_game.hpp"
#include "local_dungeon_view.hpp"
#include "mediator.hpp"
#include "menu_screen.hpp"
#include "misc/my_exceptions.hpp"
#include "options.hpp"
#include "remote_dungeon_view.hpp"
#include "server_config.hpp"
#include "task_manager.hpp"
#include "x_centre.hpp"

#include "coercri/gfx/rectangle.hpp"

unsigned int InGameScreen::getUpdateInterval()
{
    return 1000 / static_cast<unsigned int>(CfgInt("fps"));
}

InGameScreen::InGameScreen(KnightsGame &kg, const MenuSelections &msel)
    : knights_game(kg), first_update(false), sent_ready_yet(false), knights_spawned_yet(false),
      pause_mode(false), window_minimized(false),
      prev_update_time(0), stage(SETUP_STAGE)
{
    if (knights_game.server) {
        // Server setup is done here (client setup is done in separate "setup stage")
        serverSetup(knights_game.server_config->getMenu(), msel);
        stage = HANDSHAKE_STAGE;
    }
}

InGameScreen::~InGameScreen()
{
    // Upon destruction of InGameScreen (end of game) we can delete the GameClient,
    // GameServer and the DungeonViews.
    game_client.reset();
    game_server.reset();
    rdv1.reset();
    ldv2.reset();
    ldv1.reset();

    if (window) {
        // Make sure the mouse pointer is shown after leaving in-game screen
        window->showMousePointer(true);
        // Remove myself from the window if needed
        window->rmWindowListener(this);
    }
}

namespace {
    // Helper routines to configure DungeonViews properly.
    
    // Note there are two versions, one based on a ServerConfig and one that sends/receives
    // the necessary information over the network (to clients).

    // [Maybe potion and skull renderers, and menu gfx, should be shared? (TODO)]

    void SetupLocalDungeonView(LocalDungeonView &ldv, ServerConfig &server_config,
                               shared_ptr<vector<pair<string, string> > > menu_strings)
    {
        ldv.setMiniMap(Mediator::instance().getMap()->getWidth(),
                       Mediator::instance().getMap()->getHeight());
        ldv.setPotionRenderer(&server_config.getPotionRenderer());
        ldv.setSkullRenderer(&server_config.getSkullRenderer());
        ldv.setMenuGfx(server_config.getMenuCentre(), server_config.getMenuEmpty(),
                       server_config.getMenuHighlight());
        ldv.setWinnerLoserGfx(server_config.getWinnerImage(), server_config.getLoserImage());
        ldv.setMenuStrings(menu_strings);
    }
    
    void SendRemoteDviewSetup(Coercri::NetworkConnection &remote, GameServer &game_server,
                              shared_ptr<vector<pair<string, string> > > menu_strings)
    {
        remote.putCInt(Mediator::instance().getMap()->getWidth());
        remote.putCInt(Mediator::instance().getMap()->getHeight());
        remote.putCInt(menu_strings->size());
        for (int i = 0; i < menu_strings->size(); ++i) {
            remote.putString((*menu_strings)[i].first);
            remote.putString((*menu_strings)[i].second);
        }
    }
    
    void ReceiveRemoteDviewSetup(LocalDungeonView &dview, Coercri::NetworkConnection &remote,
                                 ClientConfig &client_config)
    {
        int w = remote.getCInt();
        int h = remote.getCInt();
        dview.setMiniMap(w, h);
        dview.setPotionRenderer(client_config.getPotionRenderer());
        dview.setSkullRenderer(client_config.getSkullRenderer());
        dview.setMenuGfx(client_config.getMenuCentre(), client_config.getMenuEmpty(),
                         client_config.getMenuHighlight());
        dview.setWinnerLoserGfx(client_config.getWinnerImage(), client_config.getLoserImage());

        const int num_pairs = remote.getCInt();
        shared_ptr<vector<pair<string, string> > > menu_strings(new vector<pair<string, string> >);
        menu_strings->reserve(num_pairs);
        for (int i = 0; i < num_pairs; ++i) {
            pair<string, string> p;
            p.first = remote.getString();
            p.second = remote.getString();
            menu_strings->push_back(p);
        }
        dview.setMenuStrings(menu_strings);
    }
}

void InGameScreen::serverSetup(const Menu &menu, const MenuSelections &msel)
{
    // Set up menu strings.
    shared_ptr<vector<pair<string, string> > > menu_strings = menu.getStrings(msel);
    
    // Create DungeonViews
    if (knights_game.remote) {
        // Server game
        ASSERT(knights_game.server_config);
        ldv1.reset(new LocalDungeonView(knights_game.getLeftController(),
                                        &knights_game.getRightController(),
                                        knights_game.getSoundManager()));
        rdv1.reset(new RemoteDungeonView(knights_game.remote, *knights_game.server_config, knights_game.getGfxManager()));
    } else {
        // Local game
        ldv1.reset(new LocalDungeonView(knights_game.getLeftController(), 0, knights_game.getSoundManager()));
        ldv2.reset(new LocalDungeonView(knights_game.getRightController(), 0, knights_game.getSoundManager()));
    }

    // Create GameServer
    vector<DungeonView*> dviews;
    if (ldv1) dviews.push_back(ldv1.get());
    if (ldv2) dviews.push_back(ldv2.get());
    if (rdv1) dviews.push_back(rdv1.get());
    game_server.reset(new GameServer(2));  // 2 players
    game_server->initialize(*knights_game.server_config, msel, dviews);

    // If we get to here, then the dungeon was successfully generated. Send out the 'M'
    // command to the remote (this closes the menu screen on the other side).
    if (knights_game.remote) {
        knights_game.remote->putUbyte('M');
        knights_game.remote->endPacket();
    }
    
    // Do local dview setup
    if (ldv1) SetupLocalDungeonView(*ldv1, *knights_game.server_config, menu_strings);
    if (ldv2) SetupLocalDungeonView(*ldv2, *knights_game.server_config, menu_strings);

    // Send out remote dview information
    if (knights_game.remote) {
        SendRemoteDviewSetup(*knights_game.remote, *game_server, menu_strings);
        knights_game.remote->endPacket();
    }
}

bool InGameScreen::tryClientSetup()
{
    // Create DungeonView if necessary
    if (!ldv1) {
        ldv1.reset(new LocalDungeonView(knights_game.getLeftController(),
                                        &knights_game.getRightController(),
                                        knights_game.getSoundManager()));
    }

    // Create GameClient if necessary
    if (!game_client) {
        // (note: GameClient ctor does not actually read anything from remote)
        game_client.reset(new GameClient(*knights_game.remote, *ldv1, *knights_game.client_config));
    }
    
    // Receive the remote dview information.
    // (This fails if there is no data available to read)
    if (!knights_game.remote->inAvail()) return false;
    ReceiveRemoteDviewSetup(*ldv1, *knights_game.remote, *knights_game.client_config);

    // We are done!
    return true;
}

void InGameScreen::draw(Coercri::GfxContext &gc)
{
    static const int ref_vp_width = CfgInt("dpy_viewport_width");
    static const int ref_vp_height = CfgInt("dpy_viewport_height");
    static const int ref_gutter = CfgInt("dpy_gutter");

    gc.clearClipRectangle();
    gc.clearScreen(Coercri::Color(0,0,0));
    if (!game_client && !game_server) {
        XCentre(gc, *knights_game.getFont(), knights_game.getFont()->getTextHeight() + 15, "STARTING GAME");
    } else {
        const int gvt = game_client? game_client->getGVT()
            : Mediator::instance().getGVT();

        // Have to compute appropriate scaled rectangles
        // NOTE: For split screen mode, assume 2 viewports, laid out horizontally.
        // For networked mode, we assume one viewport only.
        
        const int dpy_width = gc.getWidth();
        const int dpy_height = gc.getHeight();
        const int ref_width = ldv2 ? ref_vp_width * 2 + ref_gutter : ref_vp_width;
        const int ref_height = ref_vp_height;

        GfxManager &gm = knights_game.getGfxManager();
        float dummy;

        const float ideal_x_scale = float(dpy_width) / float(ref_width);
        const float ideal_y_scale = float(dpy_height) / float(ref_height);
        
        float x_scale, y_scale;
        gm.getGfxResizer()->roundScaleFactor(ideal_x_scale, x_scale, dummy);
        gm.getGfxResizer()->roundScaleFactor(ideal_y_scale, y_scale, dummy);
        const float scale = min(x_scale, y_scale);
        
        const int x_bias = int(ref_width * (ideal_x_scale - scale)) / 2;
        const int y_bias = int(ref_height * (ideal_y_scale - scale)) / 2;
        const int scaled_vp_width = int(ref_vp_width * scale);
        const int scaled_vp_height = int(ref_vp_height * scale);
        const int scaled_gutter = int(ref_gutter * scale);

        // work out whether we want the pause screen.
        LocalDungeonView::PauseDisplay pd = LocalDungeonView::PD_OFF;
        if (pause_mode) {
            if (ldv1 && ldv2) pd = LocalDungeonView::PD_OPAQUE;  // network game
            else pd = LocalDungeonView::PD_TRANSPARENT;          // local game
        }
        
        if (ldv1) {
            Coercri::Rectangle r(x_bias, y_bias, scaled_vp_width, scaled_vp_height);
            gc.setClipRectangle(r);
            ldv1->draw(gc, x_bias, y_bias, gm, scale, gvt, pd);
        }
        if (ldv2) {
            const int left = x_bias + scaled_vp_width + scaled_gutter;
            Coercri::Rectangle r(left, y_bias, scaled_vp_width, scaled_vp_height);
            gc.setClipRectangle(r);
            ldv2->draw(gc, left, y_bias, gm, scale, gvt, pd);
        }
        gc.clearClipRectangle();
    }
}

void InGameScreen::update()
{
    int c;

    if (knights_game.remote && knights_game.remote->getState() == Coercri::NetworkConnection::CLOSED) {
        // Hmm, connection has been lost. Go back to title
        // screen via a "Connection lost" screen.
        auto_ptr<Screen> error_screen(new ErrorScreen("The network connection has been lost."));
        knights_game.requestScreenChange(error_screen);
        return;
    }
    
    switch (stage) {
    case SETUP_STAGE:
        ASSERT(!knights_game.server);
        if (tryClientSetup()) {
            stage = HANDSHAKE_STAGE;
        }
        break;
        
    case HANDSHAKE_STAGE:
        // At this point we do a crude synchronization between the two machines. This is
        // mainly to prevent the server from starting before the client is ready.
        
        if (!knights_game.remote) {
            // Don't need this for local games:
            stage = INGAME_STAGE;
            break;
        }
        
        // Send 'ready'
        if (!sent_ready_yet) {
            knights_game.remote->putUbyte('R');
            knights_game.remote->endPacket();
            sent_ready_yet = true;
        }
        
        // Wait for 'ready'
        if (!knights_game.remote->inAvail()) break;
        c = knights_game.remote->getUbyte();
        if (c != int('R')) {
            auto_ptr<Screen> error_screen(new ErrorScreen("Protocol error"));
            knights_game.requestScreenChange(error_screen);
            return;
        }
        
        // OK, both sides are now roughly synchronized (up to lag delays etc).
        // Proceed to In-Game stage.
        stage = INGAME_STAGE;
        break;

    case INGAME_STAGE:
        {
            // Spawn knights if necessary
            if (!knights_spawned_yet && game_server) {
                game_server->spawnPlayers();
                knights_spawned_yet = true;
            }
            
            // Check for 'quit' commands on the client
            // (i.e., the server has told the client to quit)
            if (game_client && game_client->isGameQuit()) {
                stage = GAMEOVER_STAGE;
                break;
            }

            // Check for a quit request 
            // (i.e., the client has told the server it wants to quit, or,
            // someone pressed ESCAPE on the server)
            if (game_server && Mediator::instance().isQuitRequested()) {
                // This will tell the server to send out QUIT messages to clients
                Mediator::instance().quitGame();
                // This will quit the game on the server
                stage = GAMEOVER_STAGE;
                break;
            }

            // Game is paused if (pause_mode or window minimized) and local game -- in which case, we skip updates.
            // NOTE: don't allow pause on network games, as not sure how to handle that yet...
            const bool local_game = !knights_game.remote;
            const bool paused = local_game && (pause_mode || window_minimized);
            if (!paused) {
                
                // Now do an actual update
                
                // NOTE: The in-game network protocol is actually quite simple:
                // (i) Server sends "dungeonview commands" to client (these are sent from RemoteDungeonView on
                // server side; on client side they are received in GameClient::update / GameClient::receiveCommands
                // and then forwarded to the client's LocalDungeonView).
                // (ii) Client sends "controls" to server - the controls are just 1 byte IDs. (These originate
                // in the client's LocalDungeonView, are transmitted from GameClient::update on client side,
                // and received in RemoteDungeonView::readController on server side.) Alternatively client may send
                // the byte 255 to acknowledge that the game has ended.
                
                const int amount = getUpdateInterval() * CfgInt("ticks_per_second") / 1000;
                
                if (game_client) {
                    game_client->update(amount);
                } else if (game_server) {
                    game_server->update(Mediator::instance().getGVT() + amount);
                }

                // force a repaint
                window->invalidateAll();                
            }
        }
        break;

    case GAMEOVER_STAGE:
        // If this is the server, we must now wait for the client to confirm that
        // it has received the "game over" message (no. 255; see also
        // RemoteDungeonView::readController and GameClient::sendGameOverMsg).
        if (game_server && knights_game.remote) {
            if (!knights_game.remote->inAvail()) break;
            c = knights_game.remote->getUbyte();
            if (c != 255) {
                // discard non-255 characters as there may still be incoming commands pending.
                break;
            }
        }

        // At this point we must go back to the menu screen.
        auto_ptr<Screen> screen(new MenuScreen);
        knights_game.requestScreenChange(screen);
        break;
    }
}

void InGameScreen::onMouseDown(int x, int y, Coercri::MouseButton b)
{
    if (b != Coercri::MB_LEFT) return;
    if (stage != INGAME_STAGE) return;
    if (game_client && game_client->isGameOver()
    || game_server && Mediator::instance().getWinningPlayer()) {
        stage = GAMEOVER_STAGE; // this will reset everything and go back to menu
    }
}

void InGameScreen::onRawKeyDown(Coercri::RawKey rk)
{
    if (stage == INGAME_STAGE && rk == Coercri::RK_ESCAPE) {
        if (game_server) {
            Mediator::instance().requestQuit();
        } else {
            ASSERT(game_client);
            game_client->sendQuitRequest();
        }
    }

    if (rk == Coercri::RK_TAB) {
        pause_mode = !pause_mode;
        window->invalidateAll();   // make sure screen gets redrawn!
    }
}

void InGameScreen::onActivate()
{
    window_minimized = false;
}

void InGameScreen::onDeactivate()
{
    window_minimized = true;
}

bool InGameScreen::start(KnightsGame &kg, shared_ptr<Coercri::Window> win, gcn::Gui &)
{
    if (kg.getOptions().fullscreen) win->showMousePointer(false);  // hide mouse pointer in fullscreen mode.
    win->addWindowListener(this);
    window = win;
    return false;
}
