/*
 * join_game_screen.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "connecting_screen.hpp"
#include "join_game_screen.hpp"
#include "knights_game.hpp"
#include "net_msgs.hpp"
#include "title_screen.hpp"

#include "gui_button.hpp"
#include "gui_centre.hpp"
#include "gui_panel.hpp"

#include "misc/cfg.hpp"

#include "coercri/network/network_driver.hpp"
#include "coercri/network/basic_socket.hpp"

#include "boost/scoped_ptr.hpp"

namespace {
    const int update_interval = 50;
}

// Reimplement ListBox::mouseClicked so that it checks for double clicks,
// and starts the connection process if one is received.
class MyListBox : public gcn::ListBox {
public:
    explicit MyListBox(KnightsGame &kg) : knights_game(kg) { }
    
    void mouseClicked(gcn::MouseEvent &mouse_event)
    {
        gcn::ListBox::mouseClicked(mouse_event);  // make sure guichan's mouseClicked gets called
        
        if (mouse_event.getButton() == gcn::MouseEvent::LEFT && mouse_event.getClickCount() == 2) {
            auto_ptr<Screen> connecting_screen(new ConnectingScreen(getListModel()->getElementAt(getSelected())));
            knights_game.requestScreenChange(connecting_screen);
        }
    }

private:
    KnightsGame &knights_game;
};

struct ServerEntry {
    string address;
    int age;   // Number of updates ago that we last received a message from this server.
};

struct OlderThan {
    explicit OlderThan(int a) : age(a) { }
    bool operator()(const ServerEntry &se) const {
        return se.age > age;
    }
    int age;
};      

class ServerList : public gcn::ListModel {
public:
    explicit ServerList(Coercri::NetworkDriver &drv);
    virtual int getNumberOfElements() { return entries.size(); }
    virtual string getElementAt(int i) { if (i < 0 || i >= entries.size()) return ""; else return entries[i].address; }

    bool doUpdate();
    
private:
    vector<ServerEntry> entries;
    shared_ptr<Coercri::BasicSocket> socket;
    int countdown;
};

ServerList::ServerList(Coercri::NetworkDriver &drv)
{
    socket = drv.createBasicSocket(CfgInt("broadcast_port"));
    countdown = 0;
}

bool ServerList::doUpdate()
{
    bool changed = false;
    
    countdown -= update_interval;
    if (countdown <= 0) {
        // Broadcast a new request.
        socket->broadcast(BROADCAST_PING_MSG);

        // Age all the entries by 1
        for (vector<ServerEntry>::iterator it = entries.begin(); it != entries.end(); ++it) {
            it->age ++;
        }

        // Reset the timer
        countdown = CfgInt("broadcast_interval");
    }

    // Listen for incoming broadcast messages
    string msg, address;
    while (socket->receive(address, msg)) {
        if (msg == BROADCAST_PONG_MSG) {

            // See if we already know about this server. If so, set its age to zero.
            bool found = false;
            for (vector<ServerEntry>::iterator it = entries.begin(); it != entries.end(); ++it) {
                if (it->address == address) {
                    it->age = 0;
                    found = true;
                    break;
                }
            }

            if (!found) {
                // add to the end of the list.
                ServerEntry e;
                e.address = address;
                e.age = 0;
                entries.push_back(e);
                changed = true;
            }
        }
    }

    // Drop any entries older than broadcast_timeout.
    OlderThan old(CfgInt("broadcast_timeout"));
    if (find_if(entries.begin(), entries.end(), old) != entries.end()) {
        entries.erase(std::remove_if(entries.begin(), entries.end(), old),
                      entries.end());
        changed = true;
    }

    return changed;
}

class JoinGameScreenImpl : public gcn::ActionListener, public gcn::SelectionListener {
public:
    JoinGameScreenImpl(KnightsGame &kg, shared_ptr<Coercri::Window> win, gcn::Gui &g);
    void action(const gcn::ActionEvent &event);
    void valueChanged(const gcn::SelectionEvent &event);
    void doUpdate();
    
private:
    KnightsGame &knights_game;
    shared_ptr<Coercri::Window> window;
    gcn::Gui &gui;
    
    ServerList server_list;
    
    scoped_ptr<GuiCentre> centre;
    scoped_ptr<GuiPanel> panel;
    scoped_ptr<gcn::Container> container;
    scoped_ptr<gcn::Label> label1, label2;
    scoped_ptr<gcn::TextField> address_field;
    scoped_ptr<gcn::Button> connect_button, cancel_button;
    scoped_ptr<gcn::ListBox> listbox;
    scoped_ptr<gcn::ScrollArea> scroll_area;
};

namespace {
    string previously_entered_address;
}

JoinGameScreenImpl::JoinGameScreenImpl(KnightsGame &kg, shared_ptr<Coercri::Window> win, gcn::Gui &g)
    : knights_game(kg), window(win), gui(g), server_list(*knights_game.net_driver)
{
    container.reset(new gcn::Container);
    container->setOpaque(false);

    const int pad = 10;
    int y = pad;

    label1.reset(new gcn::Label("Available LAN Games (double-click to connect):"));
    label2.reset(new gcn::Label("Or enter an address to connect to:"));

    const int width = std::max(label1->getWidth(), label2->getWidth());
    
    container->add(label1.get(), pad, y);
    y += label1->getHeight() + pad;

    listbox.reset(new MyListBox(knights_game));
    listbox->setListModel(&server_list);
    const int scrollbar_width = 16;
    listbox->setWidth(width);
    listbox->addActionListener(this);
    listbox->addSelectionListener(this);
    scroll_area.reset(new gcn::ScrollArea);
    scroll_area->setContent(listbox.get());
    scroll_area->setBackgroundColor(gcn::Color(255,255,255));
    scroll_area->setOpaque(true);
    scroll_area->setSize(width, 140);
    scroll_area->setScrollbarWidth(scrollbar_width);
    scroll_area->setHorizontalScrollPolicy(gcn::ScrollArea::SHOW_NEVER);
    scroll_area->setVerticalScrollPolicy(gcn::ScrollArea::SHOW_AUTO);
    scroll_area->setFrameSize(1);
    container->add(scroll_area.get(), pad, y);
    y += scroll_area->getHeight() + pad*2;
    
    container->add(label2.get(), pad, y);
    y += label2->getHeight() + pad;

    address_field.reset(new gcn::TextField);
    address_field->adjustSize();
    address_field->setWidth(width);
    address_field->addActionListener(this);
    container->add(address_field.get(), pad, y);
    y += address_field->getHeight() + pad;
    address_field->setText(previously_entered_address);

    cancel_button.reset(new GuiButton("Cancel"));
    cancel_button->addActionListener(this);
    connect_button.reset(new GuiButton("Connect"));
    connect_button->addActionListener(this);
    container->add(connect_button.get(), pad, y);
    container->add(cancel_button.get(), pad + width - cancel_button->getWidth(), y);

    container->setSize(2*pad + width, y + cancel_button->getHeight() + pad);

    panel.reset(new GuiPanel(container.get()));
    centre.reset(new GuiCentre(panel.get()));
    gui.setTop(centre.get());

    // make sure the text field is focused initially
    address_field->requestFocus();
}

void JoinGameScreenImpl::action(const gcn::ActionEvent &event)
{
    if (event.getSource() == cancel_button.get()) {
        // Go back to title screen
        auto_ptr<Screen> title_screen(new TitleScreen);
        knights_game.requestScreenChange(title_screen);
    } else if (event.getSource() == connect_button.get() || event.getSource() == address_field.get()) {
        // Go to connecting screen
        previously_entered_address = address_field->getText();
        auto_ptr<Screen> connecting_screen(new ConnectingScreen(address_field->getText()));
        knights_game.requestScreenChange(connecting_screen);
    }
}

void JoinGameScreenImpl::valueChanged(const gcn::SelectionEvent &event)
{
    address_field->setText(server_list.getElementAt(listbox->getSelected()));
    address_field->logic();
    window->invalidateAll();  // TODO: should only invalidate the address field
}

void JoinGameScreenImpl::doUpdate()
{
    const bool changed = server_list.doUpdate();
    if (changed) {
        gui.logic();
        window->invalidateAll();  // TODO: should only invalidate the scrollarea
    }
}

bool JoinGameScreen::start(KnightsGame &kg, shared_ptr<Coercri::Window> win, gcn::Gui &gui)
{
    pimpl.reset(new JoinGameScreenImpl(kg, win, gui));
    return true;
}

void JoinGameScreen::update()
{
    pimpl->doUpdate();
}

unsigned int JoinGameScreen::getUpdateInterval()
{
    return update_interval;
}
