/*
 * loading_screen.cpp
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#include "misc.hpp"

#include "cfg.hpp"
#include "client_config.hpp"
#include "error_screen.hpp"
#include "knights_game.hpp"
#include "loading_screen.hpp"
#include "menu_screen.hpp"
#include "rstream/rstream.hpp"
#include "server_config.hpp"
#include "sound_manager.hpp"
#include "x_centre.hpp"

#ifndef DISABLE_MULTITHREADING
#include "my_exceptions.hpp"
#endif

namespace {
    void SoundCallback(const string &name, void *cxt)
    {
        SoundManager *sm = static_cast<SoundManager*>(cxt);
        sm->loadSound(name);
    }
    
    shared_ptr<ServerConfig> LoadServerConfig(GfxManager &gm, SoundManager &sm)
    {
        shared_ptr<ServerConfig> server_config(new ServerConfig(gm));
        server_config->loadConfigFile("main.txt", &SoundCallback, &sm);
        return server_config;
    }
    
    void SendServerConfig(Coercri::NetworkConnection &remote, shared_ptr<ServerConfig> server_config)
    {
        server_config->serializeAll(remote);
        remote.endPacket();
    }

    shared_ptr<ClientConfig> LoadClientConfig(Coercri::NetworkConnection &remote, GfxManager &gm)
    {
        if (!remote.inAvail()) {
            return shared_ptr<ClientConfig>();
        } else {
            shared_ptr<ClientConfig> client_config(new ClientConfig(remote, gm));
            return client_config;
        }
    }

#ifndef DISABLE_MULTITHREADING
    int LoadServerConfigThread(void *cxt)
    {
        LoadingScreen::ThreadContext & context = *static_cast<LoadingScreen::ThreadContext*>(cxt);

        try {
            shared_ptr<ServerConfig> result = LoadServerConfig(context.gm, context.sm);

            SDL_LockMutex(context.mutex);
            context.result = result;
            SDL_UnlockMutex(context.mutex);

            // We don't want exceptions to escape from the sub-thread
            // so use this error/error_message mechanism instead.
        } catch (std::exception &e) {
            SDL_LockMutex(context.mutex);
            context.error = true;
            context.error_message = e.what();
            SDL_UnlockMutex(context.mutex);
            
        } catch (...) {
            SDL_LockMutex(context.mutex);
            context.error = true;
            context.error_message = "Unknown Error";
            SDL_UnlockMutex(context.mutex);
        }

        return 0;
    }
#endif
}

bool LoadingScreen::start(KnightsGame &kg, shared_ptr<Coercri::Window> win, gcn::Gui &gui)
{
    knights_game = &kg;
    window = win;

#ifndef DISABLE_MULTITHREADING
    my_thread = 0;
    my_context = 0;
#endif
    
    return false;
}

LoadingScreen::~LoadingScreen()
{
#ifndef DISABLE_MULTITHREADING
    if (my_thread) {
        // This shouldn't really happen, but just in case:
        SDL_WaitThread(my_thread, 0);
        my_thread = 0;
    }
    delete my_context;
#endif
}

void LoadingScreen::update()
{
    if (knights_game->remote && knights_game->remote->getState() == Coercri::NetworkConnection::CLOSED) {
        // Hmm, we seem to have lost the connection. Abandon
        // load attempt and go back to title screen (via error screen).
        auto_ptr<Screen> error_screen(new ErrorScreen("The network connection has been lost."));
        knights_game->requestScreenChange(error_screen);
        return;
    }

    if (knights_game->server) {

#ifdef DISABLE_MULTITHREADING
        // Load the server config in one go.
        knights_game->server_config = LoadServerConfig(knights_game->getGfxManager(), knights_game->getSoundManager());
#else
        if (!my_context) {
            // Start a sub-thread to load the server config
            my_context = new ThreadContext(knights_game->getGfxManager(), knights_game->getSoundManager());
            my_thread = SDL_CreateThread(LoadServerConfigThread, my_context);
            return; // Allow the sub-thread to run for a bit
        } else {
            // The thread is running. Check the result
            SDL_LockMutex(my_context->mutex);
            shared_ptr<ServerConfig> result = my_context->result;
            bool error = my_context->error;
            const string msg = my_context->error_message;
            SDL_UnlockMutex(my_context->mutex);
            if (error) {
                throw InitError(msg);
            } else if (result) {
                knights_game->server_config = result;
            } else {
                return;  // thread is still running
            }
        }
#endif
        // At this point can send the server config to the client.
        // (Don't bother multi threading this as it should be quite fast, we are only buffering it
        // up for later sending by the network library).
        if (knights_game->remote) {
            SendServerConfig(*knights_game->remote, knights_game->server_config);
        }
    } else {
        ASSERT(knights_game->remote);
        knights_game->client_config = LoadClientConfig(*knights_game->remote, knights_game->getGfxManager());
        if (!knights_game->client_config) {
            // The client config has not yet been loaded. We should wait a bit and then
            // retry (at next update point).
            return;
        }
    }

    // If we get to here then we are ready to proceed to the MenuScreen....
    auto_ptr<Screen> menu_screen(new MenuScreen);
    knights_game->requestScreenChange(menu_screen);
}

void LoadingScreen::draw(Coercri::GfxContext &gc)
{
    gc.clearScreen(Coercri::Color(0,0,0));
    XCentre(gc, *knights_game->getFont(), knights_game->getFont()->getTextHeight() + 15, "LOADING");
}
