/*
 * menu.hpp
 *
 * Dungeon Environment Customization Menu.
 *
 * Copyright (c) Stephen Thompson, 2008.
 * Licensed for non-commercial use only. See LICENCE.txt for details.
 *
 */

#ifndef MENU_HPP
#define MENU_HPP

#include "coercri/gfx/graphic.hpp"

#include "boost/shared_ptr.hpp"
using namespace boost;

#include <string>
#include <vector>
using namespace std;

class DungeonDirective;
class MenuConstraint;
class MenuItem;
class MenuSelections;
class MenuValue;
class Quest;


//
// Menu -- describes the layout of the menu itself. Contains MenuItems plus some
// DungeonDirectives that are always active.
//

class Menu {
public:
    // NB contained items, directives are NOT deleted upon destruction

    void addItem(MenuItem *m)
        { if (m) items.push_back(m); }
    void addDungeonDirective(const DungeonDirective *d)
        { if (d) dungeon_directives.push_back(d); }
    void addSpaceAfter(const string &id);
    
    int getNumItems() const { return items.size(); }
    const MenuItem & getItem(int n) const { return *(items.at(n)); }
    
    int getNumDungeonDirectives() const
        { return dungeon_directives.size(); }
    const DungeonDirective * getDungeonDirective(int n) const
        { return dungeon_directives.at(n); }

    bool isSettingAllowed(const MenuSelections &msel, const string &id, int v) const;

    // Gets a vector of menu strings.
    // First string = title, second string = selected option.
    // For "isSpaceAfter" entries a pair of blank strings will be inserted.
    shared_ptr<vector<pair<string, string> > > getStrings(const MenuSelections &msel) const;
    
private:
    vector<MenuItem *> items;
    vector<const DungeonDirective *> dungeon_directives;
};


//
// MenuItem -- a particular item in the menu. Contains a list of MenuValues, of which
// one is selected at any given time.
//

class MenuItem {
public:
    MenuItem() : min_value(0), space_after(false) { } 

    void setID(const string &i) { id = i; }
    void setTitle(const string &h) { title = h; }
    void setMinValue(int n) { min_value = n; }
    void addValue(const MenuValue *m) { if (m) values.push_back(m); }
    void setSpaceAfter(bool s) { space_after = s; }
    
    const string &getID() const { return id; }
    const string &getTitle() const { return title; }
    int getMinValue() const { return min_value; }
    int getNumValues() const { return values.size(); }
    const MenuValue & getValue(int n) const { return *(values.at(n - min_value)); }
    bool getSpaceAfter() const { return space_after; }
    
private:
    int min_value;
    string id;
    string title;
    vector<const MenuValue *> values;
    bool space_after;
};


//
// MenuValue -- a possible selection for a MenuItem. Contains DungeonDirectives & other
// config info to be used when this value is selected.
//

class MenuValue {
public:
    MenuValue() : c_icon(0), g_icon(0) { }
    
    void setText(const string &s) { text = s; }
    void setCharIcon(char c) { c_icon = c; }
    void setGraphicIcon(const Coercri::Graphic *g) { g_icon = g; }
    void addConstraint(shared_ptr<MenuConstraint> c) { if (c) constraints.push_back(c); }
    void addDungeonDirective(const DungeonDirective *d)
        { if (d) dungeon_directives.push_back(d); }
    void addQuest(shared_ptr<Quest> q) { if (q) quests.push_back(q); }
    
    const string & getText() const { return text; }
    char getCharIcon() const { return c_icon; }
    const Coercri::Graphic * getGraphicIcon() const { return g_icon; }

    int getNumDungeonDirectives() const { return dungeon_directives.size(); }
    const DungeonDirective *getDungeonDirective(int n) const
        { return dungeon_directives.at(n); }
    int getNumQuests() const { return quests.size(); }
    shared_ptr<Quest> getQuest(int n) const { return quests.at(n); }

    // apply this value's constraints
    // returns true if something was changed.
    bool applyConstraints(MenuSelections &msel) const;

    // check if, when this value is selected, item id can be set to value v
    bool canSetTo(const MenuSelections &msel, const string &id, int v) const;
    
private:
    string text;
    char c_icon;
    const Coercri::Graphic *g_icon;
    vector<const DungeonDirective *> dungeon_directives;
    vector<shared_ptr<Quest> > quests;
    vector<shared_ptr<MenuConstraint> > constraints;
};

#endif
